<?php
Namespace App\Model\Admin;

use Symfony\Component\Translation\TranslatorInterface;
Use App\Utils\Utils;

/**
 * Provides admin functions for Homegear update management
 *
 */
class HomegearUpdate
{
    public $hg;
    private $utils;
    private $translator;
    private $componentUpdates = 'Updates';

    public function __construct(TranslatorInterface $translator, Utils $utils)
    {
        $this->hg = new \Homegear\Homegear;
        $this->utils = $utils;
        $this->translator = $translator;
    }

    public function getLastUpdateCheck()
    {
        return $this->hg->getData($this->componentUpdates,'lastUpdateCheck');
    }

    /**
     * Manually check for updates
     * 
     * @return array()
     */
    public function check()
    {
        if ($this->utils->checkManagementRunning(false)) {
            $this->hg->setData($this->componentUpdates,'lastUpdateCheck',time());
            $smarthaMgmt = false;
            $updateAvailable = false;
            $id = $this->hg->managementAptUpdate();
            $result = $this->utils->getCommandStatus($id);
            if ($result['exitCode'] == 0) {
                $updates['System'] = $this->hg->managementSystemUpdateAvailable();
                if ($this->utils->checkSmarthaManagementRunning(false)) {
                    $updates['Homegear'] = $this->hg->smarthaHomegearUpdateAvailable();
                    $smarthaMgmt = true;
                } else {
                    $updates['Homegear'] = $this->hg->managementHomegearUpdateAvailable();
                }
                if ($smarthaMgmt === false) {
                    // this is needed because managementHomegearUpdateAvailable() currently checks for updates of the homegear main-package only!
                    $packages = $this->_getUpdateList('homegear');
                    if (isset($packages) && is_array($packages) && count($packages) > 0) {
                        $updates['Homegear'] = true;
                    }
                }
                foreach ($updates as $type => $value) {
                    if ($value && $value != '') {
                        if ($smarthaMgmt === false && $type == 'Homegear') {
                            // homegear update ... get all updateable packages w/ versions
                            #$packages = $this->_getUpdateList('homegear');
                            // now we should have all packages to be updated along with their versions
                            if (isset($packages) && is_array($packages) && count($packages) > 0) {
                                $modulesValid = true;
                                $dependencies = true;
                                $version = str_replace('Homegear ', '', $this->hg->getVersion());
                                if (array_key_exists('homegear', $packages)) {
                                    // ok, homegear should be updated ... lets check if ALL installed modules are within the package list, too
                                    if ($this->_checkModules($packages) === false) {
                                        // there is something wrong with the modules, so don't display available update for homegear!
                                        continue;
                                    }
                                    // we're still here, so set the new version
                                    $version = $packages['homegear'];
                                }
                                // simulate update
                                $missingPackages = $this->_simulateUpdate($packages);
                                if (isset($missingPackages) && is_array($missingPackages) && count($missingPackages) > 0) {
                                    // we have packages which could not be installed, let's check dependencies to see if there are older versions to update
                                    if ($this->_checkDependencies($version,$missingPackages) === false) {
                                        // there is something wrong with dependencies, so don't display available update for homegear!
                                        #dump("false");
                                        continue;
                                    }
                                }
                                #dump("true");
                            }
                        }
                        $this->hg->setData($this->componentUpdates,'UpdateAvailable_'.$type,true);
                        $updateAvailable = true;
                    } else {
                        $this->hg->deleteData($this->componentUpdates,'UpdateAvailable_'.$type);
                    }
                }
                return ['result' => 'success', 'msg' => (!$updateAvailable ? $this->translator->trans("updates.notfound",array(),'admin') : $this->translator->trans("updates.found",array(),'admin') )];
            }
            return ['result' => 'error', 'msg' => $this->translator->trans('updates.error_checking',array(),'admin')];
        }
        return ['result' => 'error', 'msg' => $this->translator->trans('management.not_running',array(),'admin')];
    }

    /**
     * List all current updates
     * 
     * @return array
     */
    public function list()
    {
        // check if an update is running
        if ($this->utils->checkUpdateRunning('_sys-update-running') === true || $this->utils->checkUpdateRunning('_os-update-running') === true) {
            $os = -1;
            $system = -1;
        } else {
            $os = $this->hg->getData($this->componentUpdates,'UpdateAvailable_System');
            $system = $this->hg->getData($this->componentUpdates,'UpdateAvailable_Homegear');
        }
        return ['system' => $system, 'os' => $os];
    }

    /**
     * Update packages
     * 
     * @return json
     */
    public function update($type = null)
    {
        if ($type === null) { return ['result' => 'error', 'msg' => 'Update-type missing']; }
        if ($type == 'sys') {
            if ($this->utils->checkManagementRunning(false)) {
                if ($this->utils->checkSmarthaManagementRunning(false)) {
                    $id = $this->hg->smarthaAptUpgrade(intval(1));
                } else {
                    $id = $this->hg->managementAptUpgrade(intval(1));
                }
                $this->hg->deleteData($this->componentUpdates,'UpdateAvailable_UI');
                $this->hg->deleteData($this->componentUpdates,'UpdateAvailable_DeviceDefinitions');
                $this->hg->deleteData($this->componentUpdates,'UpdateAvailable_Nodes');
                $this->hg->deleteData($this->componentUpdates,'UpdateAvailable_Homegear');
                $this->hg->setData($this->componentUpdates,'UpdateRunning',$id);
                $this->utils->setOled("Update in progress...");
                return ['result' => 'success', 'type' => $type, 'id' => $id];
            }
        } elseif ($type == 'os') {
            if ($this->utils->checkManagementRunning(false)) {
                if ($this->utils->checkSmarthaManagementRunning(false)) {
                    $id = $this->hg->smarthaAptUpgrade(intval(0));
                } else {
                    $id = $this->hg->managementAptUpgrade(intval(0));
                }
                //$this->utils->saveToSession(['_os-update-running' => $id]);
                $this->hg->setData($this->componentUpdates,'UpdateRunning',$id);
                $this->utils->setOled("Update in progress...");
                return ['result' => 'success', 'type' => $type, 'id' => $id];
            }
        }
        return ['result' => 'error', 'msg' => 'Error getting update'];
    }

    private function _getUpdateList($package = false)
    {
        if ($package === false) { return; }
        $packages = array();
        $result = explode("\n",shell_exec('apt list --upgradable 2>/dev/null | grep '.$package));
        if (isset($result) && is_array($result)) {
            foreach ($result as $line) {
                $tmp = explode(" ", $line);
                if (is_array($tmp) && array_key_exists(1,$tmp)) {
                    $packages[explode("/", $tmp[0])[0]] = $tmp[1];
                }
            }
        }
        return $packages;
    }

    private function _checkModules($list = false)
    {
        if ($list === false) { return false; }
        $myModules = $this->hg->listModules();
        foreach ($myModules as $module) {
            if ($module['FAMILY_NAME'] == 'Miscellaneous') { continue; }
            // get package name of the module
            $result = explode(':',shell_exec('dpkg -S /var/lib/homegear/modules/'.$module['FILENAME']));
            if (isset($result) && is_array($result)) {
                if (!array_key_exists($result[0], $list)) {
                    // if one installed package is NOT within the update-list, the whole update may fail
                    return false;
                }
            } else {
                // uups, no package found?
                return false;
            }
        }
        return true;
    }

    private function _simulateUpdate($packages = false)
    {
        if ($packages === false) { return false; }
        $success = array();
        $result = explode("\n",shell_exec('apt-get -s upgrade |grep "Conf " |grep homegear'));
        foreach ($result as $line) {
            if (isset($line) && $line != '') {
                $line = explode(":",preg_replace('/^.*(homegear.*) \((\d.\d.\d-\d+).*/', '${1}:${2}', $line));
                $success[$line[0]] = $line[1];
            }
        }
        // return packages which could not be installed during dry-run
        return array_diff($packages,$success);
    }

    private function _checkDependencies($version = false, $list = false)
    {
        if ($version === false) { return false; }
        if ($list === false) {
            // we've to no list to check, so get all installed packages for homegear
            $installedPackages = explode("\n",shell_exec('apt list --installed |grep homegear'));
            foreach ($installedPackages as $p) {
                $p = explode("/",$p);
                if (is_array($p) && array_key_exists(0,$p) && $p[0] != '') {
                    $list[$p[0]] = '';
                }
            }
        }
        foreach ($list as $package => $v) {
            $dependencyOK = false;
            if ($package == 'homegear' || $package == 'libhomegear-base' || $package == 'homegear-nodes-core' || $package == 'homegear-nodes-extra') { continue; }
            $installedVersion = shell_exec('dpkg-query -W -f \'${version}\' '.$package);
            $result = explode("\n",shell_exec('apt-cache madison '.$package));
            foreach ($result as $line) {
                if ($line != '') {
                    $line = explode('|',$line);
                    if ($installedVersion == $line[1]) {
                        // we have reached the installed version, no need to go further ...
                        break;
                    }
                    #dump('apt-cache show '.$package.'='.trim($line[1]).' |grep Depends');
                    $dependency = shell_exec('apt-cache show '.$package.'='.trim($line[1]).' |grep Depends');
                    if (isset($dependency) && $dependency != null && $dependency != '') {
                        $neededVersion = preg_replace('/^.* homegear \((.* \d.\d.\d-\d+)\), .*\n$/', '${1}', $dependency);
                        if (strpos($neededVersion,"Depends") !== false) {
                            $neededVersion = preg_replace('/^.*, libhomegear-base \((.* \d.\d.\d-\d+)\), .*\n$/', '${1}', $dependency);
                        }
                        if ($neededVersion != '' && strpos($neededVersion,"Depends") === false) {
                            $op = explode(' ',$neededVersion);
                            list($ver,$revision) = explode('-', $op[1]);
                            $current = explode('-', $version);
                            if ($ver == $current[0]) {
                                if (trim($op[0]) == "=") {
                                    if (intval($current[1]) == intval($revision)) {
                                        $dependencyOK = true;
                                        #dump("$package: ".$op[0]." $ver $revision");
                                        break;
                                    }
                                } elseif (trim($op[0]) == ">=") {
                                    if (intval($current[1]) >= intval($revision)) {
                                        $dependencyOK = true;
                                        #dump("$package: ".$op[0]." $ver $revision");
                                        break;
                                    }
                                } elseif (trim($op[0]) == "<=") {
                                    if (intval($current[1]) <= intval($revision)) {
                                        $dependencyOK = true;
                                        #dump("$package: ".$op[0]." $ver $revision");
                                        break;
                                    }
                                }
                            }
                        } else {
                            // no homegear-dependencies found ...
                            $dependencyOK = true;
                            break;
                        }
                    } else {
                        // no dependencies found at all ...
                        $dependencyOK = true;
                        break;
                    }
                }
            }
            if ($dependencyOK === false) {
                // we have a unmet dependency, that could break the whole update!
                return false;
            }
        }
        return true;
    }
}